'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = undefined;

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _set = require('babel-runtime/core-js/set');

var _set2 = _interopRequireDefault(_set);

var _neoBlockchainNetwork = require('neo-blockchain-network');

var _neoBlockchainCore = require('neo-blockchain-core');

var _neoBlockchainNodeCore = require('neo-blockchain-node-core');

var _bloomFilter = require('bloom-filter');

var _bloomFilter2 = _interopRequireDefault(_bloomFilter);

var _bloem = require('bloem');

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _net = require('net');

var _net2 = _interopRequireDefault(_net);

var _Command = require('./Command');

var _payload = require('./payload');

var _Message = require('./Message');

var _Message2 = _interopRequireDefault(_Message);

var _errors = require('./errors');

require('./PeerData');

var _package = require('../package.json');

var _package2 = _interopRequireDefault(_package);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const createPeerBloomFilter = ({
  filter,
  k,
  tweak
}) => new _bloomFilter2.default({
  vData: Buffer.from(filter),
  nHashFuncs: k,
  nTweak: tweak
});


const createScalingBloomFilter = () => new _bloem.ScalingBloem(0.05, {
  initial_capacity: 100000,
  scaling: 4
});

const MEM_POOL_SIZE = 30000;
const GET_ADDR_PEER_COUNT = 200;
const GET_BLOCKS_COUNT = 500;
// Assume that we get 500 back, but if not, at least request every 10 seconds
const GET_BLOCKS_BUFFER = GET_BLOCKS_COUNT / 3;
const GET_BLOCKS_TIME_MS = 10000;
const GET_BLOCKS_THROTTLE_MS = 500;
const GET_BLOCKS_CLOSE_COUNT = 3;
const LOCAL_HOST_ADDRESSES = new _set2.default(['0.0.0.0', 'localhost', '127.0.0.1']);

let Node = class Node {

  constructor(options) {
    var _this = this;

    this._negotiate = (() => {
      var _ref = (0, _asyncToGenerator3.default)(function* (peer) {
        _this._sendMessage(peer, _this._createMessage({
          command: _Command.COMMAND.VERSION,
          payload: new _payload.VersionPayload({
            protocolVersion: 0,
            services: _payload.SERVICES.NODE_NETWORK,
            timestamp: Math.round(Date.now() / 1000),
            port: _this._externalPort,
            nonce: _this._nonce,
            userAgent: _this._userAgent,
            startHeight: _this._blockchain.currentBlockIndex,
            relay: true
          })
        }));

        const message = yield peer.receiveMessage(30000);
        let versionPayload;
        if (message.value.command === _Command.COMMAND.VERSION) {
          versionPayload = message.value.payload;
        } else {
          throw new _errors.NegotiationError(message);
        }

        if (!_this._checkVersion(versionPayload)) {
          throw new _errors.NegotiationError(message);
        }

        let address;
        const { host } = (0, _neoBlockchainNodeCore.getEndpointConfig)(peer.endpoint);
        if (_net2.default.isIPv4(host)) {
          address = new _payload.NetworkAddress({
            host,
            port: versionPayload.port,
            timestamp: versionPayload.timestamp,
            services: versionPayload.services
          });
        }

        _this._sendMessage(peer, _this._createMessage({ command: _Command.COMMAND.VERACK }));

        const nextMessage = yield peer.receiveMessage(30000);
        if (nextMessage.value.command !== _Command.COMMAND.VERACK) {
          throw new _errors.NegotiationError(nextMessage);
        }

        return {
          data: {
            nonce: versionPayload.nonce,
            startHeight: versionPayload.startHeight,
            bloomFilter: null,
            address
          },
          relay: versionPayload.relay
        };
      });

      return function (_x) {
        return _ref.apply(this, arguments);
      };
    })();

    this._onEvent = event => {
      if (event.event === 'PEER_CONNECT_SUCCESS') {
        const { connectedPeer } = event.extra;
        if (this._bestPeer == null || this._bestPeer.data.startHeight < connectedPeer.data.startHeight) {
          this._bestPeer = connectedPeer;
          this._requestBlocks();
        }
      } else if (event.event === 'PEER_CLOSED') {
        if (this._bestPeer === event.extra.peer) {
          this._bestPeer = this._findBestPeer();
          this._requestBlocks();
        }
      }

      this._blockchain.log((0, _extends3.default)({
        event: event.event
      }, event.data || {}));
    };

    this._requestBlocks = _lodash2.default.throttle(_lodash2.default.debounce(() => {
      const peer = this._bestPeer;
      const block = this._blockchain.currentBlock;
      if (peer != null && block.index < peer.data.startHeight) {
        if (this._getBlocksRequestsCount > GET_BLOCKS_CLOSE_COUNT) {
          peer.close();
          this._getBlocksRequestsCount = 0;
        } else if (this._shouldRequestBlocks()) {
          if (this._getBlocksRequestsIndex === block.index) {
            this._getBlocksRequestsCount += 1;
          } else {
            this._getBlocksRequestsCount = 1;
            this._getBlocksRequestsIndex = block.index;
          }
          this._getBlocksRequestTime = Date.now();
          this._sendMessage(peer, this._createMessage({
            command: _Command.COMMAND.GET_BLOCKS,
            payload: new _payload.GetBlocksPayload({
              hashStart: [block.hash]
            })
          }));
        }

        this._requestBlocks();
      }
    }, GET_BLOCKS_THROTTLE_MS), GET_BLOCKS_THROTTLE_MS);

    this._onRequestEndpoints = () => {
      this._relay(this._createMessage({ command: _Command.COMMAND.GET_ADDR }));
    };

    this._blockchain = options.blockchain;
    this._network = new _neoBlockchainNetwork.Network({
      seeds: options.seeds,
      listenTCP: options.listenTCP,
      externalEndpoints: options.externalEndpoints,
      connectPeersDelayMS: options.connectPeersDelayMS,
      maxConnectedPeers: options.maxConnectedPeers,
      socketTimeoutMS: options.socketTimeoutMS,
      negotiate: this._negotiate,
      createMessageTransform: () => new _Message.MessageTransform(this._blockchain.deserializeWireContext),
      onMessageReceived: this._onMessageReceived.bind(this),
      onRequestEndpoints: this._onRequestEndpoints.bind(this),
      onEvent: this._onEvent
    });

    this._started = false;
    this._stopped = false;

    this._externalPort = (options.listenTCP || {}).port || 0;
    this._nonce = Math.floor(Math.random() * _neoBlockchainCore.utils.UINT_MAX_NUMBER);
    this._userAgent = `NEO:neo-blockchain-js:${_package2.default.version}`;

    this.memPool = {};
    this._knownBlockHashes = createScalingBloomFilter();
    this._tempKnownBlockHashes = new _set2.default();
    this._knownTransactionHashes = createScalingBloomFilter();
    this._tempKnownTransactionHashes = new _set2.default();
    this._knownHeaderHashes = createScalingBloomFilter();
    this._tempKnownHeaderHashes = new _set2.default();
    this._getBlocksRequestsIndex = null;
    this._getBlocksRequestTime = null;
    this._getBlocksRequestsCount = 1;
  }

  get connectedPeersCount() {
    return this._network.connectedPeers.length;
  }

  start() {
    if (this._started) {
      return;
    }
    this._started = true;
    this._stopped = false;

    this._blockchain.log({ event: 'NODE_START' });
    try {
      this._network.start();
    } catch (error) {
      this._started = false;
      throw error;
    }
  }

  stop() {
    if (this._stopped) {
      return;
    }
    this._stopped = true;

    this._blockchain.log({ event: 'NODE_STOP' });
    try {
      this._network.stop();
      this._started = false;
    } catch (error) {
      this._stopped = false;
      throw error;
    }
  }

  relayTransaction(transaction) {
    var _this2 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      if (transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.MINER || _this2.memPool[transaction.hashHex] != null || _this2._tempKnownTransactionHashes.has(transaction.hashHex)) {
        return;
      }

      if (!_this2._knownTransactionHashes.has(transaction.hash)) {
        const hash = _neoBlockchainCore.common.uInt256ToString(transaction.hash);
        _this2._blockchain.log({
          event: 'RELAY_TRANSACTION_START',
          level: 'debug',
          hash
        });

        _this2._tempKnownTransactionHashes.add(transaction.hashHex);

        try {
          const foundTransaction = yield _this2._blockchain.transaction.tryGet({
            hash: transaction.hash
          });
          if (foundTransaction == null) {
            yield _this2._blockchain.verifyTransaction({
              transaction,
              memPool: _neoBlockchainCore.utils.values(_this2.memPool)
            });
            _this2.memPool[transaction.hashHex] = transaction;
            _this2._knownTransactionHashes.add(transaction.hash);
            _this2._relayTransaction(transaction);
            _this2._blockchain.log({
              event: 'RELAY_TRANSACTION_SUCCESS',
              level: 'debug',
              hash
            });
            yield _this2._trimMemPool();
          }
        } catch (error) {
          _this2._blockchain.log({ event: 'RELAY_TRANSACTION_ERROR', hash, error });

          throw error;
          // eslint-disable-next-line
        } finally {
          _this2._tempKnownTransactionHashes.delete(transaction.hashHex);
        }
      }
    })();
  }

  _relay(message) {
    this._blockchain.log({
      event: 'RELAY_MESSAGE',
      level: 'debug',
      command: message.value.command
    });
    this._network.relay(message.serializeWire());
  }

  _relayTransaction(transaction) {
    const message = this._createMessage({
      command: _Command.COMMAND.INV,
      payload: new _payload.InvPayload({
        type: _payload.INVENTORY_TYPE.TRANSACTION,
        hashes: [transaction.hash]
      })
    });
    const messagePayload = message.serializeWire();
    for (const peer of this._network.connectedPeers) {
      if (peer.relay && this._testFilter(peer.data.bloomFilter, transaction)) {
        peer.write(messagePayload);
      }
    }
  }

  _sendMessage(peer, message) {
    this._blockchain.log({
      event: 'SEND_MESSAGE',
      level: 'debug',
      command: message.value.command,
      endpoint: peer.endpoint
    });
    peer.write(message.serializeWire());
  }

  _findBestPeer() {
    return _lodash2.default.maxBy(this._network.connectedPeers, peer => peer.data.startHeight);
  }

  _shouldRequestBlocks() {
    const block = this._blockchain.currentBlock;
    const getBlocksRequestTime = this._getBlocksRequestTime;
    return this._getBlocksRequestsIndex == null || block.index - this._getBlocksRequestsIndex > GET_BLOCKS_BUFFER || getBlocksRequestTime == null || Date.now() - getBlocksRequestTime > GET_BLOCKS_TIME_MS;
  }

  _checkVersion(version) {
    if (version.nonce === this._nonce) {
      return false;
    }

    if (this._network.connectedPeers.some(peer => version.nonce === peer.data.nonce)) {
      return false;
    }

    return true;
  }

  _onMessageReceived(peer, message) {
    this._blockchain.log({
      event: 'MESSAGE_RECEIVED',
      level: 'debug',
      endpoint: peer.endpoint,
      command: message.value.command
    });

    const onError = error => {
      this._blockchain.log({
        event: 'MESSAGE_RECEIVED_ERROR',
        error,
        endpoint: peer.endpoint,
        command: message.value.command
      });
    };

    try {
      switch (message.value.command) {
        case _Command.COMMAND.ADDR:
          this._onAddrMessageReceived(message.value.payload);
          break;
        case _Command.COMMAND.BLOCK:
          this._onBlockMessageReceived(message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.CONSENSUS:
          this._onConsensusMessageReceived(message.value.payload);
          break;
        case _Command.COMMAND.FILTER_ADD:
          this._onFilterAddMessageReceived(peer, message.value.payload);
          break;
        case _Command.COMMAND.FILTER_CLEAR:
          this._onFilterClearMessageReceived(peer);
          break;
        case _Command.COMMAND.FILTER_LOAD:
          this._onFilterLoadMessageReceived(peer, message.value.payload);
          break;
        case _Command.COMMAND.GET_ADDR:
          this._onGetAddrMessageReceived(peer);
          break;
        case _Command.COMMAND.GET_BLOCKS:
          this._onGetBlocksMessageReceived(peer, message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.GET_DATA:
          this._onGetDataMessageReceived(peer, message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.GET_HEADERS:
          this._onGetHeadersMessageReceived(peer, message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.HEADERS:
          this._onHeadersMessageReceived(peer, message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.INV:
          this._onInvMessageReceived(peer, message.value.payload);
          break;
        case _Command.COMMAND.MEMPOOL:
          this._onMemPoolMessageReceived(peer);
          break;
        case _Command.COMMAND.TX:
          this._onTransactionReceived(message.value.payload).catch(onError);
          break;
        case _Command.COMMAND.VERACK:
          this._onVerackMessageReceived(peer);
          break;
        case _Command.COMMAND.VERSION:
          this._onVersionMessageReceived(peer);
          break;
        case _Command.COMMAND.ALERT:
          break;
        case _Command.COMMAND.MERKLE_BLOCK:
          break;
        case _Command.COMMAND.NOT_FOUND:
          break;
        case _Command.COMMAND.PING:
          break;
        case _Command.COMMAND.PONG:
          break;
        case _Command.COMMAND.REJECT:
          break;
        default:
          // eslint-disable-next-line
          message.value.command;
          break;
      }
    } catch (error) {
      onError(error);
    }
  }

  _onAddrMessageReceived(addr) {
    addr.addresses.filter(address => !LOCAL_HOST_ADDRESSES.has(address.host)).map(address => (0, _neoBlockchainNodeCore.createEndpoint)({
      type: 'tcp',
      host: address.host,
      port: address.port
    })).forEach(endpoint => this._network.addEndpoint(endpoint));
  }

  _onBlockMessageReceived(block) {
    var _this3 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      if (_this3._blockchain.currentBlockIndex >= block.index || _this3._tempKnownBlockHashes.has(block.hashHex)) {
        return;
      }

      if (!_this3._knownBlockHashes.has(block.hash)) {
        _this3._tempKnownBlockHashes.add(block.hashHex);

        try {
          const foundBlock = yield _this3._blockchain.block.tryGet({
            hashOrIndex: block.hash
          });
          if (foundBlock == null) {
            yield _this3._blockchain.persistBlock({ block });

            const peer = _this3._bestPeer;
            if (peer != null && block.index > peer.data.startHeight) {
              _this3._relay(_this3._createMessage({
                command: _Command.COMMAND.INV,
                payload: new _payload.InvPayload({
                  type: _payload.INVENTORY_TYPE.BLOCK,
                  hashes: [block.hash]
                })
              }));
            }
          }

          _this3._knownBlockHashes.add(block.hash);
          _this3._knownHeaderHashes.add(block.hash);
          for (const transaction of block.transactions) {
            delete _this3.memPool[transaction.hashHex];
            _this3._knownTransactionHashes.add(transaction.hash);
          }
        } finally {
          _this3._tempKnownBlockHashes.delete(block.hashHex);
        }
      }
    })();
  }

  // TODO: Implement
  // eslint-disable-next-line
  _onConsensusMessageReceived(consensus) {}

  // eslint-disable-next-line
  _onFilterAddMessageReceived(peer, filterAdd) {
    if (peer.data.bloomFilter != null) {
      peer.data.bloomFilter.insert(filterAdd.data);
    }
  }

  // eslint-disable-next-line
  _onFilterClearMessageReceived(peer) {
    // eslint-disable-next-line
    peer.data.bloomFilter = null;
  }

  // eslint-disable-next-line
  _onFilterLoadMessageReceived(peer, filterLoad) {
    // eslint-disable-next-line
    peer.data.bloomFilter = createPeerBloomFilter(filterLoad);
  }

  _onGetAddrMessageReceived(peer) {
    const addresses = _lodash2.default.take(_lodash2.default.shuffle(this._network.connectedPeers.map(connectedPeer => connectedPeer.data.address).filter(Boolean)), GET_ADDR_PEER_COUNT);
    if (addresses.length > 0) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.ADDR,
        payload: new _payload.AddrPayload({ addresses })
      }));
    }
  }

  _onGetBlocksMessageReceived(peer, getBlocks) {
    var _this4 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      const headers = yield _this4._getHeaders(getBlocks, _this4._blockchain.currentBlockIndex);
      _this4._sendMessage(peer, _this4._createMessage({
        command: _Command.COMMAND.INV,
        payload: new _payload.InvPayload({
          type: _payload.INVENTORY_TYPE.BLOCK,
          hashes: headers.map(function (header) {
            return header.hash;
          })
        })
      }));
    })();
  }

  _onGetDataMessageReceived(peer, getData) {
    var _this5 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      switch (getData.type) {
        case 0x01:
          // Transaction
          yield _promise2.default.all(getData.hashes.map((() => {
            var _ref2 = (0, _asyncToGenerator3.default)(function* (hash) {
              let transaction = _this5.memPool[_neoBlockchainCore.common.uInt256ToHex(hash)];
              if (transaction == null) {
                transaction = yield _this5._blockchain.transaction.tryGet({ hash });
              }

              if (transaction != null) {
                _this5._sendMessage(peer, _this5._createMessage({
                  command: _Command.COMMAND.TX,
                  payload: transaction
                }));
              }
            });

            return function (_x2) {
              return _ref2.apply(this, arguments);
            };
          })()));
          break;
        case 0x02:
          // Block
          yield _promise2.default.all(getData.hashes.map((() => {
            var _ref3 = (0, _asyncToGenerator3.default)(function* (hash) {
              const block = yield _this5._blockchain.block.tryGet({
                hashOrIndex: hash
              });
              if (block != null) {
                if (peer.data.bloomFilter == null) {
                  _this5._sendMessage(peer, _this5._createMessage({
                    command: _Command.COMMAND.BLOCK,
                    payload: block
                  }));
                } else {
                  _this5._sendMessage(peer, _this5._createMessage({
                    command: _Command.COMMAND.MERKLE_BLOCK,
                    payload: _this5._createMerkleBlockPayload({
                      block,
                      flags: block.transactions.map(function (transaction) {
                        return _this5._testFilter(peer.data.bloomFilter, transaction);
                      })
                    })
                  }));
                }
              }
            });

            return function (_x3) {
              return _ref3.apply(this, arguments);
            };
          })()));
          break;
        case 0xe0:
          // Consensus
          // TODO: Implement
          break;
        default:
          // eslint-disable-next-line
          getData.type;
          break;
      }
    })();
  }

  _onGetHeadersMessageReceived(peer, getBlocks) {
    var _this6 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      const headers = yield _this6._getHeaders(getBlocks, _this6._blockchain.currentHeader.index);
      _this6._sendMessage(peer, _this6._createMessage({
        command: _Command.COMMAND.HEADERS,
        payload: new _payload.HeadersPayload({ headers })
      }));
    })();
  }

  _onHeadersMessageReceived(peer, headersPayload) {
    var _this7 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      const headers = headersPayload.headers.filter(function (header) {
        return !_this7._knownHeaderHashes.has(header.hash) && !_this7._tempKnownHeaderHashes.has(header.hashHex);
      });
      if (headers.length > 0) {
        for (const header of headers) {
          _this7._tempKnownHeaderHashes.add(header.hashHex);
        }
        try {
          yield _this7._blockchain.persistHeaders(headers);
          for (const header of headers) {
            _this7._knownHeaderHashes.add(header.hash);
          }
        } finally {
          for (const header of headers) {
            _this7._tempKnownHeaderHashes.delete(header.hashHex);
          }
        }
      }

      if (_this7._blockchain.currentHeader.index < peer.data.startHeight) {
        _this7._sendMessage(peer, _this7._createMessage({
          command: _Command.COMMAND.GET_HEADERS,
          payload: new _payload.GetBlocksPayload({
            hashStart: [_this7._blockchain.currentHeader.hash]
          })
        }));
      }
    })();
  }

  _onInvMessageReceived(peer, inv) {
    let hashes;
    switch (inv.type) {
      case 0x01:
        // Transaction
        hashes = inv.hashes.filter(hash => !this._knownTransactionHashes.has(hash) && !this._tempKnownTransactionHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
        break;
      case 0x02:
        // Block
        hashes = inv.hashes.filter(hash => !this._knownBlockHashes.has(hash) && !this._tempKnownBlockHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
        break;
      case 0xe0:
        // Consensus
        // TODO: Implement
        hashes = [];
        break;
      default:
        // eslint-disable-next-line
        inv.type;
        hashes = [];
        break;
    }

    if (hashes.length > 0) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.GET_DATA,
        payload: new _payload.InvPayload({ type: inv.type, hashes })
      }));
    }
  }

  _onMemPoolMessageReceived(peer) {
    this._sendMessage(peer, this._createMessage({
      command: _Command.COMMAND.INV,
      payload: new _payload.InvPayload({
        type: _payload.INVENTORY_TYPE.TRANSACTION,
        hashes: _neoBlockchainCore.utils.values(this.memPool).map(transaction => transaction.hash)
      })
    }));
  }

  _onTransactionReceived(transaction) {
    var _this8 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      yield _this8.relayTransaction(transaction);
    })();
  }

  // eslint-disable-next-line
  _onVerackMessageReceived(peer) {
    peer.close();
  }

  // eslint-disable-next-line
  _onVersionMessageReceived(peer) {
    peer.close();
  }

  _getHeaders(getBlocks, maxHeight) {
    var _this9 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      let hashStopIndexPromise = _promise2.default.resolve(maxHeight);
      if (getBlocks.hashStop !== 0) {
        hashStopIndexPromise = _this9._blockchain.header.tryGet({ hashOrIndex: getBlocks.hashStop }).then(function (hashStopHeader) {
          return hashStopHeader == null ? maxHeight : Math.min(hashStopHeader.index, maxHeight);
        });
      }
      const [hashStartHeaders, hashEnd] = yield _promise2.default.all([_promise2.default.all(getBlocks.hashStart.map(function (hash) {
        return _this9._blockchain.header.tryGet({ hashOrIndex: hash });
      })), hashStopIndexPromise]);
      const hashStartHeader = _lodash2.default.head(_lodash2.default.orderBy(hashStartHeaders.filter(Boolean), [function (header) {
        return header.index;
      }]));
      if (hashStartHeader == null) {
        return [];
      }
      const hashStart = hashStartHeader.index + 1;
      if (hashStart > maxHeight) {
        return [];
      }

      const headers = yield _promise2.default.all(_lodash2.default.range(hashStart, Math.min(hashStart + GET_BLOCKS_COUNT, hashEnd)).map(function (index) {
        return _this9._blockchain.header.get({ hashOrIndex: index });
      }));

      return headers;
    })();
  }

  _trimMemPool() {
    var _this10 = this;

    return (0, _asyncToGenerator3.default)(function* () {
      const memPool = _neoBlockchainCore.utils.values(_this10.memPool);
      if (memPool.length > MEM_POOL_SIZE) {
        const transactionAndFee = yield _promise2.default.all(memPool.map((() => {
          var _ref4 = (0, _asyncToGenerator3.default)(function* (transaction) {
            const networkFee = yield transaction.getNetworkFee({
              getOutput: _this10._blockchain.output.get,
              governingToken: _this10._blockchain.settings.governingToken,
              utilityToken: _this10._blockchain.settings.utilityToken,
              fees: _this10._blockchain.settings.fees
            });
            return [transaction, networkFee];
          });

          return function (_x4) {
            return _ref4.apply(this, arguments);
          };
        })()));
        const hashesToRemove = _lodash2.default.take(_lodash2.default.sortBy(transactionAndFee,
        // TODO: Might be a bug since we're converting to number
        function ([transaction, networkFee]) {
          return networkFee.divn(transaction.size).toNumber();
        })
        // eslint-disable-next-line
        ).map(function ([transaction, _]) {
          return transaction.hashHex;
        });
        for (const hash of hashesToRemove) {
          delete _this10.memPool[hash];
        }
      }
    })();
  }

  _testFilter(bloomFilterIn, transaction) {
    const bloomFilter = bloomFilterIn;
    if (bloomFilter == null) {
      return true;
    }
    return bloomFilter.contains(transaction.hash) || transaction.outputs.some(output => bloomFilter.contains(output.address)) || transaction.inputs.some(input => bloomFilter.contains(input.serializeWire())) || transaction.scripts.some(script => bloomFilter.contains(_neoBlockchainCore.crypto.toScriptHash(script.verification))) || transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.REGISTER && transaction instanceof _neoBlockchainCore.RegisterTransaction && bloomFilter.contains(transaction.asset.admin);
  }

  _createMerkleBlockPayload({
    block,
    flags
  }) {
    const tree = new _neoBlockchainCore.MerkleTree(block.transactions.map(transaction => transaction.hash)).trim(flags);

    const buffer = Buffer.allocUnsafe(Math.floor((flags.length + 7) / 8));
    for (let i = 0; i < flags.length; i += 1) {
      if (flags[i]) {
        // eslint-disable-next-line
        buffer[Math.floor(i / 8)] |= 1 << i % 8;
      }
    }

    return new _payload.MerkleBlockPayload({
      version: block.version,
      previousHash: block.previousHash,
      merkleRoot: block.merkleRoot,
      timestamp: block.timestamp,
      index: block.index,
      consensusData: block.consensusData,
      nextConsensus: block.nextConsensus,
      script: block.script,
      transactionCount: block.transactions.length,
      hashes: tree.toHashArray(),
      flags: buffer
    });
  }

  _createMessage(value) {
    return new _Message2.default({
      magic: this._blockchain.settings.messageMagic,
      value
    });
  }
};
exports.default = Node;